from laegr.featurization.Dump2Features import Dump2Features, Graph
from ovito.data import NearestNeighborFinder
from typing import Union, Any


class GraphFeatures(Dump2Features):
    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)

    def to_graph(self, N: int) -> Graph:
        nodes = self.get_graph_node()
        edge_index, edge_attr = self.get_edge(N)
        labels = self.get_labels()
        graph_features = Graph(nodes, edge_index, edge_attr, None, labels)
        return graph_features

    def get_graph_node(self) -> list[list[Union[float, int]]]:
        raise NotImplementedError

    def get_edge(self, N: int) -> tuple[list[list[int | Any]], list[list[float]]]:
        finder = NearestNeighborFinder(N, self.data)
        edge_index = []
        edge_attr = []
        for index in range(self.data.particles.count):
            for neigh in finder.find(index):
                edge_index.append([index, neigh.index])
                edge_attr.append(self.get_edge_attr(neigh))
        return edge_index, edge_attr

    @staticmethod
    def get_edge_attr(neigh: NearestNeighborFinder) -> list[float]:
        raise NotImplementedError


class AtomicNumber(GraphFeatures):
    def get_graph_node(self) -> list[list[int]]:
        ptypes = self.data.particles.particle_types
        nodes = []
        transform = {1: 42, 2: 41, 3: 73, 4: 74}
        for atom_type in ptypes:
            cur_node = transform[atom_type]
            nodes.append([cur_node])
        return nodes


class AtomicNumberDistance(AtomicNumber):
    @staticmethod
    def get_edge_attr(neigh: NearestNeighborFinder) -> list[float]:
        return [neigh.distance]


class AtomicNumberPos(AtomicNumber):
    def to_graph(self) -> Graph:
        nodes = self.get_graph_node()
        pos = self.get_pos()
        labels = self.get_labels()
        graph_features = Graph(nodes, None, None, pos, labels)
        return graph_features

    def get_pos(self) -> list[list[float]]:
        pos_list = []
        for cur_pos in self.data.particles.positions:
            pos_list.append(list(cur_pos))
        return pos_list


class AtomicMass(GraphFeatures):
    def get_graph_node(self) -> list[list[float]]:
        nodes = self.data.particles['Mass'][...]
        nodes = [[i] for i in nodes]
        return nodes


class AtomicMassDistance(AtomicMass):
    @staticmethod
    def get_edge_attr(neigh: NearestNeighborFinder) -> list[float]:
        return [neigh.distance]


class AtomicMass3D(AtomicMass):
    @staticmethod
    def get_edge_attr(neigh: NearestNeighborFinder) -> list[float]:
        return neigh.delta


class OneHotGraph(GraphFeatures):
    def get_graph_node(self) -> list[list[int]]:
        ptypes = self.data.particles.particle_types
        nodes = []
        for atom_type in ptypes:
            cur_node = self.to_onehot(atom_type)
            nodes.append(cur_node)
        return nodes

    @staticmethod
    def to_onehot(atom_type: int) -> list[int]:
        raise NotImplementedError


class OneHotDistance(OneHotGraph):
    @staticmethod
    def get_edge_attr(neigh: NearestNeighborFinder) -> list[float]:
        return [neigh.distance]


class OneHot3D(OneHotGraph):
    @staticmethod
    def get_edge_attr(neigh: NearestNeighborFinder) -> list[float]:
        return neigh.delta


class OneHotDistance_TypeOrder(OneHotDistance):
    @staticmethod
    def to_onehot(atom_type: int) -> list[int]:
        feature = [0, 0, 0, 0]
        feature[atom_type - 1] = 1
        return feature


class OneHotDistance_AtomicOrder(OneHotDistance):
    @staticmethod
    def to_onehot(atom_type: int) -> list[int]:
        if atom_type == 1:
            new_atom_type = 2
        elif atom_type == 2:
            new_atom_type = 1
        else:
            new_atom_type = atom_type
        feature = [0, 0, 0, 0]
        feature[new_atom_type - 1] = 1
        return feature


class OneHot3D_TypeOrder(OneHot3D):
    @staticmethod
    def to_onehot(atom_type: int) -> list[int]:
        feature = [0, 0, 0, 0]
        feature[atom_type - 1] = 1
        return feature


class OneHot3D_AtomicOrder(OneHot3D):
    @staticmethod
    def to_onehot(atom_type: int) -> list[int]:
        if atom_type == 1:
            new_atom_type = 2
        elif atom_type == 2:
            new_atom_type = 1
        else:
            new_atom_type = atom_type
        feature = [0, 0, 0, 0]
        feature[new_atom_type - 1] = 1
        return feature



from typing import Tuple
import torch
from torch_geometric.loader import DataLoader
import numpy as np
from tqdm import tqdm

class TrainingValidation:
    def __init__(self, model: torch.nn.Module, loss_fn = torch.nn.MSELoss(), optimizer=torch.optim.Adam, lr=0.001, epoch=5000) -> None:
        self.model = model
        self.loss_fn = loss_fn
        self.optimizer = optimizer(self.model.parameters(), lr=lr)
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.model = self.model.to(self.device)
        self.epoch = epoch
    
    def train(self, data: DataLoader, clip=False) -> float:
        losses = []
        for batch in tqdm(data):
            batch.to(self.device)
            self.optimizer.zero_grad()
            # pred = self.model(batch.x, batch.edge_index, batch.edge_attr, batch.batch)
            pred = self.model_prediction(batch)
            loss = self.loss_fn(pred, batch.y)
            loss.backward()
            if clip:
                clipping_value = 10 # arbitrary value of your choosing
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), clipping_value)
            self.optimizer.step()
            losses.append(np.sqrt(loss.item()))
        return np.mean(losses), np.max(losses), np.min(losses), np.std(losses)
            
    def validation(self, data: DataLoader) -> Tuple[list[float], list[float], list[float]]:
        y_P, y_T, losses = [], [], []
        with torch.no_grad():
            for batch in tqdm(data):
                batch.to(self.device)
                # pred = self.model(batch.x, batch.edge_index, batch.edge_attr, batch.batch)
                pred = self.model_prediction(batch)
                y_P += pred.tolist()
                y_T += batch.y.tolist()
                loss = torch.nn.L1Loss()(batch.y, pred)
                loss = loss.item()
                losses.append(loss)
        return y_T, y_P, np.mean(losses)
    
    def model_prediction(self, batch: DataLoader) -> torch.tensor:
        pred = self.model(batch.x, batch.edge_index, batch.edge_attr)
        return pred
    
    def savingModel(self, path: str) -> None:
        torch.save(self.model.state_dict(), path)


class TrainingValidation_SchNet(TrainingValidation):
    def model_prediction(self, batch: DataLoader) -> torch.tensor:
        # print(batch)
        batch.batch = torch.tensor([i for i in range(len(batch.z))], dtype=torch.long)
        # print(batch.batch)
        batch.to(self.device)
        pred = self.model(batch.z, batch.pos, batch.batch)
        # print(pred)
        return pred


    


from typing import Union, Optional
from ovito.io import import_file
import pandas as pd


class Graph:
    def __init__(self, nodes: list[list[Union[float, int]]], edge_index: Optional[list[list[int]]],
                 edge_attr: Optional[list[list[float]]], pos: Optional[list[list[float]]],
                 labels: list[list[float]]
                 ) -> None:
        self.nodes = nodes
        self.edge_index = edge_index
        self.edge_attr = edge_attr
        self.pos = pos
        self.labels = labels


class Dump2Features:
    """
    Base class to generate the Graph/Matrix (node features, edges and labels/PE) of the given dump file
    The dump file should contain the information of atomic mass and atomic potential energy
    """

    def __init__(self, file_path: str) -> None:
        pipeline = import_file(file_path)
        self.data = pipeline.compute()

    def __toDataFrame(self) -> pd:
        df = pd.DataFrame([self.data.particles.identifiers, self.data.particles.particle_types,
                           self.data.particles['Mass'][...], self.data.particles['c_eng'][...]]).T
        df.columns = ["id", 'type', 'mass', 'c_eng']
        return df

    def get_node_features(self) -> NotImplemented:
        raise NotImplementedError

    def get_labels(self) -> list[list[float]]:
        labels = self.data.particles['c_eng'][...]
        labels = [[i] for i in labels]
        return labels



import os, re, pickle
from tqdm import tqdm
from torch_geometric.data.data import Data
import torch


from laegr.featurization.Dump2Features import Graph
from laegr.featurization.GraphFeatures import AtomicMassDistance, AtomicMass3D, AtomicNumberDistance, AtomicNumberPos
from laegr.featurization.GraphFeatures import OneHotDistance_TypeOrder, OneHotDistance_AtomicOrder, OneHot3D_TypeOrder, \
    OneHot3D_AtomicOrder


class DataSaving:
    def __init__(self, sourcePath: str, outputDir: str, filer=1000) -> None:
        self.files = self.__read_data(sourcePath, filter=filer)
        self.outputDir = outputDir

    def save_data(self) -> None:
        raise NotImplementedError

    def to_features(self, dumpPath: str) -> Graph:
        raise NotImplementedError

    def __read_data(self, read_path: str, filter=1) -> list[str]:
        if os.path.isdir(read_path):
            return self.__read_data_dir(read_path, filter)
        else:
            return self.__read_data_single(read_path)

    @staticmethod
    def __read_data_dir(data_path: str, filter: int) -> list[str]:
        files = os.listdir(data_path)
        files = [f for f in files if ("mc" in f and "dump" in f)]
        files.sort(key=lambda var: [int(x) if x.isdigit() else x for x in re.findall(r'[^0-9]|[0-9]+', var)])
        L = []
        for f in files:
            [x] = re.findall('[0-9]+', f)
            x = int(x)
            if x % filter == 0:
                L.append(os.path.join(data_path, f))
        return L

    @staticmethod
    def __read_data_single(data_path: str) -> list[str]:
        return [data_path]


class GNN_Data(DataSaving):
    def __init__(self, sourcePath: str, outputDir: str, num: int, feature: str, filer=1000) -> None:
        super().__init__(sourcePath, outputDir, filer)
        self.num = num
        self.feature = feature

    def saveData(self) -> None:
        output = []
        for file in tqdm(self.files):
            graph = self.to_features(file)
            data = self.to_GNN_data(graph)
            output.append(data)
        open_file = open(self.outputDir, "wb")
        pickle.dump(output, open_file)
        open_file.close()

    def to_features(self, dumpPath: str) -> Graph:
        match self.feature:
            case "AtomicMassDistance":
                graph = AtomicMassDistance(dumpPath).to_graph(self.num)
            case "AtomicMass3D":
                graph = AtomicMass3D(dumpPath).to_graph(self.num)
            case "AtomicNumberDistance":
                graph = AtomicNumberDistance(dumpPath).to_graph(self.num)
            case "AtomicNumberPos":
                graph = AtomicNumberPos(dumpPath).to_graph()
            case "OneHotDistance_TypeOrder":
                graph = OneHotDistance_TypeOrder(dumpPath).to_graph(self.num)
            case "OneHotDistance_AtomicOrder":
                graph = OneHotDistance_AtomicOrder(dumpPath).to_graph(self.num)
            case "OneHot3D_TypeOrder":
                graph = OneHot3D_TypeOrder(dumpPath).to_graph(self.num)
            case "OneHot3D_AtomicOrder":
                graph = OneHot3D_AtomicOrder(dumpPath).to_graph(self.num)
            case _:
                raise Exception('''Only support eight different features:   
                                        AtomicMassDistance,             AtomicMass3D,
                                        AtomicNumberDistance,           AtomicNumberPos,
                                        OneHotDistance_TypeOrder,       OneHotDistance_AtomicOrder,
                                        OneHot3D_TypeOrder,             OneHot3D_AtomicOrder.''')
        return graph

    @staticmethod
    def to_GNN_data(graph: Graph) -> Data:
        labels = graph.labels
        if graph.labels:
            labels = torch.tensor(graph.labels, dtype=torch.float)
        if graph.pos:
            nodes = torch.tensor(graph.nodes, dtype=torch.long)
            edge_pos = torch.tensor(graph.pos, dtype=torch.float)
            data = Data(z=nodes, edge_attr=None, pos=edge_pos, y=labels)
        else:
            nodes = torch.tensor(graph.nodes, dtype=torch.float)
            edge_index = torch.tensor(graph.edge_index, dtype=torch.long)
            edge_attr = torch.tensor(graph.edge_attr, dtype=torch.float)
            data = Data(x=nodes, edge_index=edge_index.t().contiguous(), edge_attr=edge_attr, y=labels)
        return data





import torch, os, torch_geometric
from torch_geometric.loader import DataLoader
from torch_geometric.nn import GraphConv, SAGEConv, GCNConv, NNConv, CGConv, SchNet
from laegr.model.gnn.Model import Model, Model_GNN_Base, Model_GNN_NN, Model_SchNet
from laegr.model.gnn.Model import Model_GNN_SchNet, Model_GNN_Crystal
from laegr.model.gnn.Layer import NNConv_Edge_Interaction, NodeCat
from laegr.model.gnn.TrainingValidation import TrainingValidation, TrainingValidation_SchNet
import numpy as np
import pandas as pd


class Train:
    def __init__(self, layer_type: str) -> None:
        self.layer_func = self.__determine_layer(layer_type)
        self.layer_type = layer_type

    def build_gnn_model(self, first_layer_size: int, embedding_size: int, layer_number: int) -> torch.nn.Module:
        self.layer_number = layer_number
        self.embedding_size = embedding_size
        gnn_model = Model_GNN_Base(first_layer_size, embedding_size, self.layer_func, layer_number).build_model()
        return gnn_model

    def build_model(self, gnn_model: torch.nn, output_model: torch.nn, activation=torch.nn.ReLU()) -> torch.nn.Module:
        model = Model(gnn_model, output_model, activation)
        return model

    @staticmethod
    def __determine_layer(layer_type: str) -> torch_geometric.nn:
        match layer_type:
            case "GraphConv":
                return GraphConv
            case "SAGEConv":
                return SAGEConv
            case "GCNConv":
                return GCNConv
            case "CGConv":
                return CGConv
            case "NNConv":
                return NNConv
            case "CGConv":
                return CGConv
            case "SchNet":
                return SchNet
            case "NNConv_Edge_Interaction":
                return NNConv_Edge_Interaction
            case "NodeCat":
                return NodeCat
            case _:
                raise Exception(f"Do not support the type: {layer_type}")

    def training(self, model: Model, train_data: DataLoader, validation_data: DataLoader, epoch=100, clip=False,
                 optimizer=torch.optim.Adam, lr=0.001, loss_fn=torch.nn.MSELoss(), training_loss_switch=True,
                 output="") -> None:
        Tra_Val = self.create_train_agent(model, epoch, optimizer, loss_fn, lr)
        training_loss, val_loss = [], []
        for cur_epoch in range(epoch):
            Tra_Val.train(train_data, clip)
            cur_train_loss = -1
            if training_loss_switch:
                _, _, cur_train_loss = Tra_Val.validation(train_data)
            _, _, cur_val_loss = Tra_Val.validation(validation_data)
            print(f"current epoch: {cur_epoch}. Training loss is: {cur_train_loss}. Validation loss is: {cur_val_loss}")
            training_loss.append(cur_train_loss)
            val_loss.append(cur_val_loss)
        outputName = self.get_output_name(output)
        pd.DataFrame([training_loss, val_loss]).T.to_csv(f"{outputName}_training_loss.csv")

    def get_output_name(self, output: str) -> str:
        outputName = f"{str(self.layer_type)}_{str(self.layer_number)}_{str(self.embedding_size)}_{output}"
        return outputName

    @staticmethod
    def create_train_agent(model, epoch, optimizer, loss_fn, lr):
        Tra_Val = TrainingValidation(model=model, epoch=epoch, optimizer=optimizer, loss_fn=loss_fn, lr=lr)
        return Tra_Val


class TrainCrystal(Train):
    def __init__(self, layer_type: str) -> None:
        super().__init__(layer_type)

    def build_gnn_model(self, first_layer_size: int, layer_number: int, edge_dim: int) -> torch.nn.Module:
        self.layer_number = layer_number
        self.embedding_size = 1
        gnn_model = Model_GNN_Crystal(first_layer_size, layer_number, edge_dim).build_model()
        return gnn_model


class TrainEdge(Train):
    def __init__(self, layer_type: str) -> None:
        super().__init__(layer_type)

    def build_gnn_model(self, first_layer_size: int, embedding_size: int, layer_number: int,
                        edge_model: torch.nn.Module) -> torch.nn.Module:
        self.layer_number = layer_number
        self.embedding_size = embedding_size
        gnn_model = Model_GNN_NN(first_layer_size, embedding_size, self.layer_func, layer_number,
                                 edge_model).build_model()
        return gnn_model


class TrainSchNet(Train):
    def build_model(self, gnn_model: torch.nn, output_model: torch.nn, activation=torch.nn.ReLU()) -> torch.nn.Module:
        model = Model_SchNet(gnn_model, output_model, activation)
        return model

    def build_gnn_model(self, layer_number: int = 1,
                        hidden_channels: int = 128, num_filters: int = 128, num_interactions: int = 6,
                        num_gaussians: int = 50, cutoff: float = 4.7, max_num_neighbors: int = 26
                        ) -> torch.nn.Module:
        self.layer_number = layer_number
        self.embedding_size = hidden_channels
        self.num_filters = num_filters
        self.num_interactions = num_interactions
        self.num_gaussians = num_gaussians
        gnn_model = Model_GNN_SchNet(layer_number).build_model(hidden_channels, num_filters, num_interactions,
                                                               num_gaussians, cutoff, max_num_neighbors)
        return gnn_model

    def get_output_name(self, output: str) -> str:
        outputName = f"{str(self.layer_type)}_{str(self.layer_number)}_{str(self.embedding_size)}_{self.num_filters}_{self.num_interactions}_{self.num_gaussians}_{output}"
        return outputName

    @staticmethod
    def create_train_agent(model, epoch, optimizer, loss_fn, lr):
        Tra_Val = TrainingValidation_SchNet(model=model, epoch=epoch, optimizer=optimizer, loss_fn=loss_fn, lr=lr)
        return Tra_Val


class TrainEdgeCombine(Train):
    def __init__(self, layer_type: str) -> None:
        super().__init__(layer_type)



import os
import pickle
from typing import Tuple, Union, Optional
from torch_geometric.loader import DataLoader
import random

class LoadData():
    def __init__(self, path: Union[str, list[str]]):
        self.__List = self.__readList(path)

    @staticmethod
    def __readList(path: Union[str, list[str]]):
        print("Start reading Data")
        listRawData = []
        if isinstance(path, str):
            print("single data")
            open_file   = open(path, "rb")
            listRawData = pickle.load(open_file)
            open_file.close()
        elif isinstance(path, list):
            print("Multiple data.")
            for cur_path in path:
                print("Read current data ...")
                cur_file = open(cur_path, 'rb')
                cur_data = pickle.load(cur_file)
                cur_file.close()
                listRawData.extend(cur_data)
        else:
            raise Exception("Please provide correct data path!")
        print("Reading data Done")
        return listRawData
    
    def transformInput(self, partition: float = 0.8, train_batch_size: int = 1, val_batch_size: int = 1, shuffle: bool = True, random_see: int = 66) -> Tuple[DataLoader, Optional[DataLoader]]:
        data_size    = len(self.__List)
        random.seed(random_see)
        if (shuffle):
            random.shuffle(self.__List)
        if partition == -1:
            train_loader = DataLoader(self.__List, batch_size=train_batch_size)
            return train_loader, None
        train_loader       = DataLoader(self.__List[:int(data_size*partition)], batch_size=train_batch_size)
        validation_loader  = DataLoader(self.__List[int(data_size*partition):], batch_size=val_batch_size)
        return train_loader, validation_loader